@@ -5,7 +5,10 @@ from urllib.parse import parse_qs
 from urllib.parse import urljoin
 from urllib.parse import urlparse
 from urllib.parse import urlsplit
+import base64
+import binascii
 import ipaddress
+import json
 import logging
 import magic
 import re
@@ -15,7 +18,7 @@ import warnings
 warnings.filterwarnings('ignore', category=UserWarning, module='bs4')
 
 
-def _tokenize(bytes, extra_tokens=True):
+def _tokenize(bytes, mimetype, extra_tokens=True):
     """ This function tokenizes the input bytes based on some common characters. It returns
         the tokens in ASCII format. """
 
@@ -48,13 +51,15 @@ def _tokenize(bytes, extra_tokens=True):
     #results = re.compile(b'\{([^\{\}]+)\}').findall(ascii_bytes)
     tokens += results
 
-    # Find anything sandwiched between ' '
-    results = re.compile(b'\'\s*(.*?)\s*\'').findall(ascii_bytes)
-    tokens += results
+    if not mimetype == 'data':
 
-    # Find anything sandwiched between " "
-    results = re.compile(b'\"\s*(.*?)\s*\"').findall(ascii_bytes)
-    tokens += results
+        # Find anything sandwiched between ' '
+        results = re.compile(b'\'\s*(.*?)\s*\'').findall(ascii_bytes)
+        tokens += results
+
+        # Find anything sandwiched between " "
+        results = re.compile(b'\"\s*(.*?)\s*\"').findall(ascii_bytes)
+        tokens += results
 
     # Find anything sandwiched between spaces
     results = re.compile(b'\ ([^\ ]+)\ ').findall(ascii_bytes)
@@ -96,15 +101,18 @@ def _tokenize(bytes, extra_tokens=True):
     return tokens
 
 
-def _ascii_find_urls(bytes, extra_tokens=True):
+def _ascii_find_urls(bytes, mimetype, extra_tokens=True):
     """ This function finds URLs inside of ASCII bytes. """
 
-    tokens = _tokenize(bytes, extra_tokens=extra_tokens)
+    tokens = _tokenize(bytes, mimetype, extra_tokens=extra_tokens)
+
+    # Remove any leading mailto: from the URLs.
+    tokens = [t[7:] if t.lower().startswith('mailto:') else t for t in tokens]
 
     return tokens
 
 
-def _html_find_urls(bytes, base_url=None):
+def _html_find_urls(bytes, mimetype, base_url=None):
     """ This function finds URLs inside of valid HTML bytes. """
 
     def _recursive_tag_values(tag, values=[]):
@@ -124,6 +132,15 @@ def _html_find_urls(bytes, base_url=None):
 
         return values
 
+    def tag_visible(element):
+        """ This sub-function gets the visible elements from the page. """
+
+        if element.parent.name in ['style', 'script', 'head', 'title', 'meta', '[document]']:
+            return False
+        if isinstance(element, Comment):
+            return False
+        return True
+
     # Only convert the ASCII bytes to HTML.
     ascii_bytes = b''.join(re.compile(b'[\x00\x09\x0A\x0D\x20-\x7E]{8,}').findall(bytes))
     ascii_bytes = ascii_bytes.replace(b'\x00', b'')
@@ -134,14 +151,40 @@ def _html_find_urls(bytes, base_url=None):
     # Convert the bytes into soup. Also try url decoding the bytes to bypass some obfuscation.
     soups = []
     soups.append(BeautifulSoup(ascii_bytes, 'lxml'))
+    soups.append(BeautifulSoup(ascii_bytes, 'html.parser'))
     try:
         soups.append(BeautifulSoup(urllib.parse.unquote(str(ascii_bytes)), 'lxml'))
     except:
         pass
+ 
+    # ugly hack for handling universal character set obfuscation like so:
+    # convert bytes -> string -> bytes will transform:
+    #  `https:/\\link\xc3\x82\xc2\xadedin.\xc3\x82\xc2\xadcom/redirect?url=`
+    # TO
+    #  `https://linkedin.com/redirect?url=`
+    try:
+        soups.append(BeautifulSoup(bytes.decode('utf-8').encode(encoding='ascii', errors='ignore'), features="lxml"))
+    except Exception as e:
+        logging.info("Encountered exception attempting universal character set hack: {}".format(e))
 
     # Loop over both soups.
     for soup in soups:
 
+        # Remove any obfuscating <font> tags that do not render in the browser.
+        # NOTE: This lambda expression "should" work, but it causes several tests to fail for some reason.
+        # font_tag_list = soup.findAll(lambda tag: tag.name == 'font' and len(tag.attrs) == 1 and tag['id'])
+        font_tag_list = soup.findAll('font', {'id' : True})
+        font_tag_list = [tag for tag in font_tag_list if len(tag.attrs) == 1]
+        visible_urls = []
+        if font_tag_list:
+            for tag in font_tag_list:
+                tag.decompose()
+
+            texts = soup.findAll(text=True)
+            visible_texts = filter(tag_visible, texts)
+            text = ''.join(t for t in visible_texts).encode('ascii', errors='ignore')
+            visible_urls = _ascii_find_urls(text, 'ascii')
+
         # Find any URLs embedded in script + document.write JavaScript.
         # \r\n\r\ndocument.write(unescape('<meta HTTP-EQUIV="REFRESH" content="0; url=http://somebadsite.com/catalog/index.php">'));\r\n\r\n
         script_urls = []
@@ -162,7 +205,11 @@ def _html_find_urls(bytes, base_url=None):
                         code = code[1:-1]
                     # Turn the string into bytes and feed it back into the _html_find_urls function.
                     code = code.encode()
-                    script_urls += _html_find_urls(code)
+                    script_urls += _html_find_urls(code, mimetype)
+
+        # Find any URLs embedded in <script> tags. Use the _ascii_find_urls function to try and catch everything.
+        for script_tag in script_tags:
+            script_urls += _ascii_find_urls(str(script_tag).encode('utf-8'), mimetype)
 
         # Find any meta-refresh URLs.
         # <meta HTTP-Equiv="refresh" content="0; URL=UntitledNotebook1.html">
@@ -198,7 +245,7 @@ def _html_find_urls(bytes, base_url=None):
         for tag in soup.find_all('base'):
             try:
                 if tag['href']:
-                    base_url = tag['href']
+                    base_url = tag['href'].replace('\\', '/')
             except:
                 pass
 
@@ -246,6 +293,7 @@ def _html_find_urls(bytes, base_url=None):
             urls += css_urls
             urls += meta_urls
             urls += script_urls
+            urls += visible_urls
 
         # As a last-ditch effort, find URLs in the visible text of the HTML. However,
         # we only want to add strings that are valid URLs as they are. What we do not
@@ -261,13 +309,26 @@ def _html_find_urls(bytes, base_url=None):
         # URL by the is_valid function.
         urls = [u.strip() for u in urls]
 
+        # Remove any leading mailto: from the URLs.
+        urls = [u[7:] if u.lower().startswith('mailto:') else u for u in urls]
+
         # Remove any leading //'s from the URLs.
         urls = [u[2:] if u.startswith('//') else u for u in urls]
 
+        # Fix cases like http:/ instead of http://
+        urls = [u.replace(':/', '://') if '://' not in u else u for u in urls]
+
+        # Remove any leading spaces or %20 from the URLs.
+        urls = [u.strip() for u in urls]
+
+        # Fix any http://\\\\ and http://\\ instances.
+        urls = [u.replace('://\\\\', '://') for u in urls]
+        urls = [u.replace('://\\', '://') for u in urls]
+
     return urls
 
 
-def _pdf_find_urls(bytes):
+def _pdf_find_urls(bytes, mimetype):
     """ This function finds URLs inside of PDF bytes. """
 
     # Start with only the ASCII bytes. Limit it to 12+ character strings.
@@ -338,7 +399,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
         # If the bytes are HTML...
         if 'html' in mimetype:
             try:
-                all_urls += _html_find_urls(thing, base_url)
+                all_urls += _html_find_urls(thing, mimetype, base_url)
             except:
                 if log:
                     logger.exception('Error when finding HTML URLs.')
@@ -346,7 +407,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
         # If the bytes are a PDF...
         elif 'pdf' in mimetype:
             try:
-                all_urls += _pdf_find_urls(thing)
+                all_urls += _pdf_find_urls(thing, mimetype)
             except:
                 if log:
                     logger.exception('Error when finding PDF URLs.')
@@ -358,19 +419,19 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
         # If the bytes are ASCII or Unicode text...
         elif 'ascii' in mimetype or 'text' in mimetype:
             try:
-                all_urls += _html_find_urls(thing, base_url)
+                all_urls += _html_find_urls(thing, mimetype, base_url)
             except:
                 if log:
                     logger.exception('Error when finding ASCII/HTML URLs.')
 
             try:
-                all_urls += _ascii_find_urls(thing)
+                all_urls += _ascii_find_urls(thing, mimetype)
             except:
                 if log:
                     logger.exception('Error when finding ASCII URLs.')
 
             try:
-                all_urls += _pdf_find_urls(thing)
+                all_urls += _pdf_find_urls(thing, mimetype)
             except:
                 if log:
                     logger.exception('Error when finding ASCII/PDF URLs.')
@@ -380,7 +441,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
 
             # Try to treat the bytes as a PDF and find URLs.
             try:
-                all_urls += _pdf_find_urls(thing)
+                all_urls += _pdf_find_urls(thing, mimetype)
             except:
                 if log:
                     logger.exception('Error when finding unknown/PDF URLs.')
@@ -389,7 +450,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
             # In that case, we don't want to find all possible ASCII URLs, as it will result
             # in a lot of bad URLs. Try to treat the bytes as ASCII and find URLs.
             try:
-                all_urls += _ascii_find_urls(thing, extra_tokens=False)
+                all_urls += _ascii_find_urls(thing, mimetype, extra_tokens=False)
             except:
                 if log:
                     logger.exception('Error when finding unknown/ASCII URLs.')
@@ -414,7 +475,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
     # and re-find any URLs that way.
     if not valid_urls and 'html' in mimetype:
         try:
-            for url in _ascii_find_urls(thing):
+            for url in _ascii_find_urls(thing, mimetype):
                 # If the URL is valid as-is, just add it to the list.
                 if is_valid(url):
                     valid_urls.append(url)
@@ -439,8 +500,10 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
         except:
             pass
 
-    # Check if any of the URLs are Proofpoint URLs and try to decode them.
+    # Try to decode various special URLs.
     for url in ascii_urls[:]:
+
+        # Check if any of the URLs are Proofpoint URLs and try to decode them.
         if 'urldefense.proofpoint.com/v2/url' in url:
             try:
                 query_u=parse_qs(urlparse(url).query)['u'][0]
@@ -451,8 +514,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
                 if log:
                     logger.exception('Error decoding Proofpoint URL: {}'.format(url))
 
-    # Check if any of the URLs are Outlook safelinks and try to decode them.
-    for url in ascii_urls[:]:
+        # Check if any of the URLs are Outlook safelinks and try to decode them.
         if 'safelinks.protection.outlook.com' in url:
             try:
                 query_url=parse_qs(urlparse(url).query)['url'][0]
@@ -463,17 +525,59 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
                 if log:
                     logger.exception('Error decoding Outlook safelinks URL: {}'.format(url))
 
-    # Check if any of the URLs are Google redirection URLs and try to decode them.
-    for url in ascii_urls[:]:
+        # Check if any of the URLs are Google redirection URLs and try to decode them.
         if 'www.google.com/url?' in url:
             try:
-                query_url=parse_qs(urlparse(url).query)['q'][0]
+                query_url=parse_qs(urlparse(url).query)
+                if 'url' in query_url:
+                    redirect_url = query_url['url'][0]
+                    redirect_url = urllib.parse.unquote(redirect_url)
+                    if is_valid(redirect_url):
+                        ascii_urls.append(redirect_url)
+                # not sure this format is even used any longer but leaving the code here
+                elif 'q' in query_url:
+                    query_url = query_url['q'][0]
+                    decoded_url = urllib.parse.unquote(query_url)
+                    if is_valid(decoded_url):
+                        ascii_urls.append(decoded_url)
+            except:
+                if log:
+                    logger.exception('Error decoding Google redirection URL: {}'.format(url))
+
+        # Check if any of the URLs are Barracuda Link Protect URLs and try to decode them.
+        if 'linkprotect.cudasvc.com' in url:
+            try:
+                query_url=parse_qs(urlparse(url).query)['a'][0]
                 decoded_url = urllib.parse.unquote(query_url)
                 if is_valid(decoded_url):
                     ascii_urls.append(decoded_url)
             except:
                 if log:
-                    logger.exception('Error decoding Google redirection URL: {}'.format(url))
+                    logger.exception('Error decoding Barracuda Link Protect URL: {}'.format(url))
+
+        # Check if any of the URLs are mandrillapp URLs and try to decode them.
+        if 'mandrillapp.com' in url and 'p=' in url:
+            try:
+                query_base64 = parse_qs(urlparse(url).query)['p'][0]
+                try:
+                    decoded = base64.b64decode(query_base64)
+                except binascii.Error:
+                    decoded = base64.b64decode(query_base64 + '==')
+                j = json.loads(decoded)
+                j = json.loads(j['p'])
+                if is_valid(j['url']):
+                    ascii_urls.append(j['url'])
+            except:
+                if log:
+                    logger.exception('Error decoding mandrillapp URL: {}'.format(url))
+
+        # Pull out URLs from linkedin.com/redirect service
+        if 'linkedin.com' in url and 'redirect' in url:
+            parsed_url = urlparse(url)
+            redirect_url_base = parse_qs(parsed_url.query)['url'][0]
+            redirect_url = redirect_url_base + '/#' + parsed_url.fragment
+            if is_valid(redirect_url):
+                ascii_urls.append(redirect_url)
 
     # Add an unquoted version of each URL to the list.
     for url in ascii_urls[:]:
@@ -483,7 +587,7 @@ def find_urls(thing, base_url=None, mimetype=None, log=False):
     # catch URLs that may not have the scheme on the front of them.
     ascii_urls = ['http://' + u if not u.lower().startswith('http') and not u.lower().startswith('ftp') else u for u in ascii_urls]
 
-    # Remove any trailing "/" from the URLs so that they are consistent with how they go in CRITS.
+    # Remove any trailing "/" from the URLs so that they are consistent with how they go into the intel database.
     ascii_urls = [u[:-1] if u.endswith('/') else u for u in ascii_urls]
 
     return sorted(list(set(ascii_urls)))
@@ -524,11 +628,6 @@ def is_valid(url, fix=True):
         if fix and not split_url.scheme:
             split_url = urlsplit('http://{}'.format(url))
 
-        # Check for the edge case of results returned by find_urls, such as google.com URLs
-        # like: http://google.com#default#userData
-        if split_url.netloc and not split_url.path and split_url.fragment:
-            return False
-
         # Check if the netloc has a ':' in it, which indicates that
         # there is a port number specified. We need to remove that in order
         # to properly check if it is a valid IP address.
@@ -557,6 +656,14 @@ def is_valid(url, fix=True):
                 split_url = urlsplit(user_pass_url)
                 netloc = split_url.netloc
 
+            # Look for the edge case of the URL having a username without password notation.
+            # The path needs to have something in it to avoid counting email addresses as URLs.
+            if ':' not in split_url.netloc and '@' in split_url.netloc and len(split_url.path) > 1:
+                user = re.compile(r'(.*?@)').findall(split_url.netloc)[0]
+                user_url = url.replace(user, '')
+                split_url = urlsplit(user_url)
+                netloc = split_url.netloc
+
             # Check the netloc. Check if it is an IP address.
             try:
                 ipaddress.ip_address(netloc)
